/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2000-2002, Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZTWIN32FASTMUTEX_H__
#define __ZTWIN32FASTMUTEX_H__

#if !defined(_WIN32)
#error "Win32 specific implementation"
#endif

#include "zthread/ZThread.h"
#include "zthread/Lockable.h"
#include <windows.h>
#include <assert.h>

namespace ZThread {


/**
 * @class FastMutex
 *
 * @author Eric Crahen <crahen@cse.buffalo.edu>
 * @date <2002-03-04T20:21:15-0500>
 * @version 2.0.0a
 *
 * This is the smallest and fastest synchronization object in the library.
 * It is an implementation of fast mutex, an all or nothing exclusive
 * lock. It should be used only where you need speed and are willing 
 * to sacrifice all the state & safety checking provided by the framework
 * for speed.
 *
 * This uses a custom spin lock based on the typical compare-and-swap 
 * approach. There are two version of these CAS functions, one was the
 * original 32-bit platforms, and the others are for the 64-bit platforms.
 * Depending on what service packs you have installed, you have access to 
 * just the 32-bit set: 
 *
 *   LONG InterlockedExchange(LPLONG, LONG)
 *   LONG InterlockedCompareExchange(LPLONG, LONG, LONG, LONG) or
 *   LPVOID InterlockedCompareExchange(LPVOID*, LPVOID, LPVOID, LPVOID) or
 *
 * or the 64-bit as well.
 *
 *   LPVOID InterlockedExchangePointer(LPVOID*, LPVOID)
 *   LPVOID InterlockedCompareExchangePointer(LPVOID*, LPVOID, LPVOID, LPVOID)
 *
 * The values are exactly the same, it just depends on whats declared and if 
 * you want 64-bit or 32-bit support.
 *
 * ----
 * Because Windows 95 and earlier can run on processors prior to the 486, they
 * don't all support the CMPXCHG function, and so Windows 95 an earlier dont support
 * InterlockedCompareExchange. If you define _WIN95, you'll get a version of the 
 * FastMutex that just uses critical sections.
 */ 
class ZTHREAD_API FastMutex : public Lockable {

// Add def for mingw32 or other non-ms compiler to align on 64-bit
// boundary
#pragma pack(push, 8)
  LONG volatile _lock;
#pragma pack(pop)
  
  public:
  
  /**
   * Create a new FastMutex
   */
  inline FastMutex() 
    throw(Synchronization_Exception) : _lock(0) { }

  
  /**
   * Destroy FastMutex
   */
  virtual ~FastMutex() {
    assert(_lock == 0);
  }

  /**
   * Lock the fast Mutex, no error check.
   * 
   * @exception None
   */ 
  inline virtual void acquire() 
    throw(Synchronization_Exception) {

#if defined (_MSC_VER) && (_MSC_VER <= 1200) 
    while (::InterlockedCompareExchange((PVOID*)&_lock, (PVOID)1, (PVOID)0) != 0)
#else
    while (::InterlockedCompareExchange((LPLONG)&_lock, 1, 0) != 0)
#endif
      ::Sleep(0);

  }

  
  /**
   * Release the fast Mutex, no error check.
   * 
   * @exception None
   */ 
  inline virtual void release() 
    throw(Synchronization_Exception) {

    ::InterlockedExchange((LPLONG)&_lock, (LONG)0);

  }
    
  /**
   * Try to acquire an exclusive lock. No safety or state checks are performed.
   * This function returns immediately regardless of the value of the timeout
   *
   * @param timeout Unused
   * @return bool
   * @exception Synchronization_Exception - not thrown
   */
 inline virtual bool tryAcquire(unsigned long)    
   throw(Synchronization_Exception) {
 
#if defined (_MSC_VER) && (_MSC_VER <= 1200) 
   return ::InterlockedCompareExchange((PVOID*)&_lock, (PVOID)1, (PVOID)0) == 0;
#else
   return ::InterlockedCompareExchange((LPLONG)&_lock, 1, 0) == 0;
#endif

 }
  
}; /* Fast Mutex */


};
#endif
