/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2000-2002, Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZTWIN32FASTRECURSIVEMUTEX_H__
#define __ZTWIN32FASTRECURSIVEMUTEX_H__

#if !defined(_WIN32)
#error "Win32 specific implementation"
#endif

#include "zthread/ZThread.h"
#include "zthread/Lockable.h"
#include <windows.h>
#include <assert.h>

namespace ZThread {


/**
 * @class FastRecursiveMutex
 *
 * @author Eric Crahen <crahen@cse.buffalo.edu>
 * @date <2002-03-04T20:21:24-0500>
 * @version 2.0.0a
 *
 * This is the smaller and faster implementation of a RecursiveMutex. 
 * A single thread can acquire the mutex any number of times, but it
 * must perform a release for each acquire(). Other threads are blocked
 * until a thread has released all of its locks on the mutex.
 * 
 * This particular implementation performs fewer safety checks. Like 
 * the FastMutex implementation, any waiting caused by an acquire() request 
 * is not interruptable. This is so that the mutex can have the fastest
 * response time for a time critical application while still having a good
 * degree of reliability.
 *
 * TryEnterCriticalSection() does not work at all on some systems, so its 
 * not used.
 *
 * Instead, a custom spin lock based on the typical compare-and-swap approach. 
 * is used. There are two version of these CAS functions, one was the
 * original 32-bit platforms, and the others are for the 64-bit platforms.
 * Depending on what service packs you have installed, you have access to 
 * just the 32-bit set: 
 *
 *   LONG InterlockedExchange(LPLONG, LONG)
 *   LONG InterlockedCompareExchange(LPLONG, LONG, LONG, LONG) or
 *   LPVOID InterlockedCompareExchange(LPVOID*, LPVOID, LPVOID, LPVOID) or
 *
 * or the 64-bit as well.
 *
 *   LPVOID InterlockedExchangePointer(LPVOID*, LPVOID)
 *   LPVOID InterlockedCompareExchangePointer(LPVOID*, LPVOID, LPVOID, LPVOID)
 *
 * The values are exactly the same, it just depends on whats declared and if 
 * you want 64-bit or 32-bit support.
 *
 * ----
 * Because Windows 95 and earlier can run on processors prior to the 486, they
 * don't all support the CMPXCHG function, and so Windows 95 an earlier dont support
 * InterlockedCompareExchange. If you define _WIN95, you'll get a version of the 
 * FastMutex that just uses critical sections.
 */ 
class ZTHREAD_API FastRecursiveMutex : public Lockable {

// Add def for mingw32 or other non-ms compiler to align on 64-bit
// boundary
#pragma pack(push, 8)
  LONG volatile _lock;
#pragma pack(pop)
  LONG _count;
  
  public:
  
  /**
   * Create a new FastRecursiveMutex
   */
  inline FastRecursiveMutex() 
    throw(Synchronization_Exception) : _lock(0), _count(0) { }

  
  /**
   * Destroy FastMutex
   */
  virtual ~FastRecursiveMutex() {
    assert(_lock == 0);
  }

  /**
   * Lock the fast Mutex, no error check.
   * 
   * @exception None
   */ 
  inline virtual void acquire() 
    throw(Synchronization_Exception) {

    DWORD id = ::GetCurrentThreadId();

    // Take ownership if the lock is free or owned by the calling thread
    do {

      DWORD owner = (DWORD)::InterlockedCompareExchange((PVOID*)&_lock, (PVOID)id, (PVOID)0);
      if(owner == 0 || owner == id)
        break;

      ::Sleep(0);

    } while(1);
        
    _count++;

  }

  
  /**
   * Release the fast Mutex, no error check.
   * 
   * @exception None
   */ 
  inline virtual void release() 
    throw(Synchronization_Exception) {

    if(--_count == 0)
      ::InterlockedExchange((LPLONG)&_lock, (LONG)0);

  }
    
  /**
   * Try to acquire an exclusive lock. No safety or state checks are performed.
   * This function returns immediately regardless of the value of the timeout
   *
   * @param timeout Unused
   * @return bool
   * @exception Synchronization_Exception - not thrown
   */
 inline virtual bool tryAcquire(unsigned long)    
   throw(Synchronization_Exception) {

   DWORD id = ::GetCurrentThreadId();
   DWORD owner = (DWORD)::InterlockedCompareExchange((PVOID*)&_lock, (PVOID)id, (PVOID)0);

   if(owner == 0 || owner == id) {
      _count++;
      return true;
   }

   return false;

 }

 
}; /* FastRecursiveMutex */


};
#endif
