/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2000-2002, Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZTWIN95FASTMUTEX_H__
#define __ZTWIN95FASTMUTEX_H__

#if !defined(HAVE_WIN95)
#error "Win95 fast mutex implementation, use Win32 implementation."
#endif

#include "zthread/ZThread.h"
#include "zthread/Lockable.h"
#include <windows.h>
#include <assert.h>

namespace ZThread {


/**
 * @class FastMutex
 *
 * @author Eric Crahen <crahen@cse.buffalo.edu>
 * @date <2002-03-02T09:05:32-0500>
 * @version 2.0.0a
 *
 * This is the smallest and fastest synchronization object in the library.
 * It is an implementation of fast mutex, an all or nothing exclusive
 * lock. It should be used only where you need speed and are willing 
 * to sacrifice all the state & safety checking provided by the framework
 * for speed.
 *
 * This uses a custom spin lock based on the older swap & compare approach
 * using the XCHG instruction. You should only use this is you *absolutely* need
 * to use an older system, like Windows 95. If you can, use the Win32 version.
 *
 * Because Windows 95 and earlier can run on processors prior to the 486, they
 * don't all support the CMPXCHG function, and so Windows 95 an earlier dont support
 * InterlockedCompareExchange.
 *
 * This is asm inlined for microsoft visual c, it needs to be changed in order to
 * compile with gcc, or another win32 compiler - but more likely than not you'll
 * be using the Win32 version on those compilers and this won't be a problem.
 */ 
class ZTHREAD_API FastMutex : public Lockable {

// Add def for mingw32 or other non-ms compiler to align on 32-bit boundary
#pragma pack(push, 4)
  unsigned char volatile _lock;
#pragma pack(pop)
  
  public:
  
  /**
   * Create a new FastMutex
   */
  inline FastMutex() 
    throw(Synchronization_Exception) : _lock(0) { }

  
  /**
   * Destroy FastMutex
   */
  virtual ~FastMutex() {
    assert(_lock == 0);
  }

  /**
   * Lock the fast Mutex, no error check.
   * 
   * @exception None
   */ 
  inline virtual void acquire() 
    throw(Synchronization_Exception) {

    DWORD dw = (DWORD)&_lock;

    _asm { // swap & compare
    spin_lock:

      mov  al, 1
      mov  esi, dw
      xchg [esi], al       
      and  al,al
      jz   spin_locked
    }

    ::Sleep(0);

    _asm {
      jmp spin_lock
    spin_locked:
    }

  }

  
  /**
   * Release the fast Mutex, no error check.
   * 
   * @exception None
   */ 
  inline virtual void release() 
    throw(Synchronization_Exception) {

    DWORD dw = (DWORD)&_lock;

    _asm { 
      mov  al, 0
      mov  esi, dw
      xchg [esi], al       
    }


  }
    
  /**
   * Try to acquire an exclusive lock. No safety or state checks are performed.
   * This function returns immediately regardless of the value of the timeout
   *
   * @param timeout Unused
   * @return bool
   * @exception Synchronization_Exception - not thrown
   */
 inline virtual bool tryAcquire(unsigned long)    
   throw(Synchronization_Exception) {

    volatile DWORD dw = (DWORD)&_lock;
    volatile DWORD result;

    _asm {

      mov  al, 1
      mov  esi, dw
      xchg [esi], al       
      and  al,al
      mov  esi, result
      xchg [esi], al       
    }

    return result == 0;

 }
  
}; /* Fast Mutex */


};
#endif
