/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2000-2002, Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZTWIN95FASTRECURSIVEMUTEX_H__
#define __ZTWIN95FASTRECURSIVEMUTEX_H__

#include "zthread/Win95_FastMutex.h"

namespace ZThread {


/**
 * @class FastRecursiveMutex
 *
 * @author Eric Crahen <crahen@cse.buffalo.edu>
 * @date <2002-03-02T09:03:56-0500>
 * @version 2.0.0a
 *
 * This is the smallest and fastest synchronization object in the library.
 * It is an implementation of fast mutex, an all or nothing exclusive
 * lock. It should be used only where you need speed and are willing 
 * to sacrifice all the state & safety checking provided by the framework
 * for speed.
 *
 * This uses a custom spin lock based on the older swap & compare approach
 * using the XCHG instruction. You should only use this is you *absolutely* need
 * to use an older system, like Windows 95. If you can, use the Win32 version.
 *
 * Because Windows 95 and earlier can run on processors prior to the 486, they
 * don't all support the CMPXCHG function, and so Windows 95 an earlier dont support
 * InterlockedCompareExchange.
 *
 * This is asm inlined for microsoft visual c, it needs to be changed in order to
 * compile with gcc, or another win32 compiler - but more likely than not you'll
 * be using the Win32 version on those compilers and this won't be a problem.
 */ 
class ZTHREAD_API FastRecursiveMutex : public Lockable {

  FastMutex _lock;
  volatile DWORD _owner;
  volatile DWORD _count;

 public:
  
  /**
   * Create a new FastRecursiveMutex
   */
  inline FastRecursiveMutex() 
    throw(Synchronization_Exception) : _owner(0), _count(0) { }

  
  /**
   * Destroy FastRecursiveMutex
   */
  virtual ~FastRecursiveMutex() {
    assert(_owner == 0);
  }

  /**
   * Lock the fast Mutex, no error check.
   * 
   * @exception None
   */ 
  inline virtual void acquire() 
    throw(Synchronization_Exception) {

    DWORD id = ::GetCurrentThreadId();

    do {

      _lock.acquire();

      // Take ownership
      if(_owner == 0) 
        _owner = id;

      // Update count
      if(_owner == id) {
        _count++;
        id = 0;
      }

      _lock.release();

    } while(id != 0);

  }

  
  /**
   * Release the fast Mutex, no error check.
   * 
   * @exception None
   */ 
  inline virtual void release() 
    throw(Synchronization_Exception) {

    DWORD id = ::GetCurrentThreadId();

    _lock.acquire();

    assert(_owner == id);
    if(--_count == 0)
      _owner = 0;

    _lock.release();

  }
    
  /**
   * Try to acquire an exclusive lock. No safety or state checks are performed.
   * This function returns immediately regardless of the value of the timeout
   *
   * @param timeout Unused
   * @return bool
   * @exception Synchronization_Exception - not thrown
   */
 inline virtual bool tryAcquire(unsigned long)    
   throw(Synchronization_Exception) {

    DWORD id = ::GetCurrentThreadId();

    _lock.acquire();

    // Take ownership
    if(_owner == 0)
       _owner = id;

    // Update count
    if(_owner == id) {
       _count++;
       id = 0;
    }

    _lock.release();

    return id == 0;

 }
  
}; /* Fast Mutex */


};
#endif
