/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001, 2002 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZTVALUEHOLDER_H__
#define __ZTVALUEHOLDER_H__

namespace ZThread {

class AbstractThreadLocal;

/**
 * @class ValueHolder
 * @author Eric Crahen <crahen@cse.buffalo.edu>
 * @date <2002-02-17T21:24:42-0500>
 *
 * A ValueHolder is used to hold values for ThreadLocals. It simplifies
 * detecting the first use of a ThreadLocal, as well as propogating values to
 * children.
 */
class ValueHolder {
  
  const AbstractThreadLocal* _tls;
  void* _value;

  public:
  
  /**
   * Create a new ValueHolder with given value.
   *
   * @param const AbstractThreadLocal*
   * @param void* value
   */
  ValueHolder(const AbstractThreadLocal* tls, void* value = 0) : _tls(tls) {
    setValue(value);
  }

  /**
   * Get the ThreadLocal associated with this ValueHolder
   *
   * @return const AbstractThreadLocal
   */
  const AbstractThreadLocal* getThreadLocal() {
    return _tls;
  }

  /**
   * Get the current value held by this object.
   *
   * @return void*
   */
  inline void* getValue() {
    return _value;
  }

  /**
   * Set the current value held by this object and return
   * the old value.
   *
   * @param void* new value
   * @return void* old value
   */
  inline void* setValue(void* value) {

    void* oldValue = _value;
    _value = value;

    return oldValue;

  }

};

};

#endif
